package org.netkernel.json.core.util;

import java.io.*;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.*;

import javax.json.Json;
import javax.json.stream.JsonGenerator;
import javax.json.stream.JsonGeneratorFactory;
import javax.json.stream.JsonParser;
import javax.json.stream.JsonParser.Event;
import javax.json.stream.JsonParserFactory;

import org.netkernel.mod.hds.HDSFactory;
import org.netkernel.mod.hds.IHDSDocument;
import org.netkernel.mod.hds.IHDSMutator;
import org.netkernel.mod.hds.IHDSNode;

public class HDS2Utils
{
	public static final String ARRAY_INDICATOR_SUFFIX="__A";
	private static JsonGeneratorFactory sFactory;
	private static JsonParserFactory sParserFactory;
	
	static
	{
		sFactory=Json.createGeneratorFactory(Collections.EMPTY_MAP);
		sParserFactory=Json.createParserFactory(Collections.EMPTY_MAP);
	}
	
	public static void serializeHDS(OutputStream aOS, IHDSDocument aHDS, boolean aRemoveRootElement) throws IOException
	{
		JsonGenerator jg = sFactory.createGenerator(aOS);
		IHDSNode root=aHDS.getRootNode();
		IHDSNode[] children=root.getChildren();
		if (children.length==1 && children[0].getName().endsWith(ARRAY_INDICATOR_SUFFIX))
		{	serializeHDSArray(children[0],jg,true);
		}
		else
		{	jg.writeStartObject();
			if (aRemoveRootElement && children.length==1)
			{	children=children[0].getChildren();
			}
			for (IHDSNode child : children)
			{	serializeHDSObject(child,jg);
			}
			jg.writeEnd();
		}
		jg.close();
	}
	
	private static void serializeHDSArray(IHDSNode aNode, JsonGenerator aJG, boolean aIsRoot)
	{
		String name=aNode.getName();
		name=name.substring(0, name.length()-ARRAY_INDICATOR_SUFFIX.length());
		String actualName=name;
		while (actualName.endsWith(ARRAY_INDICATOR_SUFFIX))
		{	actualName=actualName.substring(0, name.length()-ARRAY_INDICATOR_SUFFIX.length());
		}
		IHDSNode[] children=aNode.getChildren();
		if (aIsRoot)
		{	aJG.writeStartArray();
		}
		else
		{	aJG.writeStartArray(actualName);
		}
		for (IHDSNode child: children)
		{	if (child.getName().equals(name))
			{	
				if (child.getName().endsWith(ARRAY_INDICATOR_SUFFIX))
				{	serializeHDSArray(child,aJG,true);
				}
				else
				{
					aJG.writeStartObject();
					for (IHDSNode grandchild: child.getChildren())
					{	serializeHDSObject(grandchild,aJG);
					}
					aJG.writeEnd();
				}
			}
			else
			{	throw new IllegalArgumentException("array elements must be named correctly");
			}	
		}
		aJG.writeEnd();
	}
	
	private static void serializeHDSObject(IHDSNode aNode, JsonGenerator aJG)
	{	String name=aNode.getName();
		IHDSNode[] children=aNode.getChildren();
		Object value=aNode.getValue();
		if (value!=null && children.length>0) throw new IllegalArgumentException("node with both value and children not supported in JSON");
		if (children.length>0 || (name.endsWith(ARRAY_INDICATOR_SUFFIX) && value==null))
		{	if (name.endsWith(ARRAY_INDICATOR_SUFFIX))
			{	serializeHDSArray(aNode,aJG,false);
			}
			else
			{	aJG.writeStartObject(name);
				for (IHDSNode child: children) serializeHDSObject(child,aJG);
				aJG.writeEnd();
			}
		}
		else
		{	if (value==null)
			{	aJG.writeNull(name);
			}
			else if (value instanceof String)
			{	aJG.write(name,(String)value);
			}
			else if (value instanceof BigDecimal)
			{	aJG.write(name,(BigDecimal)value);
			}
			else if (value instanceof BigInteger)
			{	aJG.write(name,(BigInteger)value);
			}
			else if (value instanceof Boolean)
			{	aJG.write(name,(Boolean)value);
			}
			else if (value instanceof Double)
			{	aJG.write(name,(Double)value);
			}
			else if (value instanceof Integer)
			{	aJG.write(name,(Integer)value);
			}
			else if (value instanceof Long)
			{	aJG.write(name,(Long)value);
			}
		}
		
		
	}
	
	public static void parseHDS(IHDSMutator aMutator, InputStream aIS, boolean aToString) throws Exception
	{	JsonParser jr = sParserFactory.createParser(aIS);
		Event event = null;
		while(jr.hasNext())
		{	event = jr.next();
			if (event==Event.START_ARRAY)
			{	parseHDSArray("hds",jr,aMutator, aToString);
			}
			else if (event==Event.START_OBJECT)
			{	parseHDSObject(null,jr,aMutator, aToString);
			}
		}
	}
	
	public static IHDSDocument parseHDS(InputStream aIS) throws Exception
	{	IHDSMutator m=HDSFactory.newDocument();
		parseHDS(m,aIS,false);
		return m.toDocument(false);
	}
	
	private static void parseHDSArray(String aName, JsonParser aJR, IHDSMutator aMutator, boolean aToString)
	{	aMutator.pushNode(aName+ARRAY_INDICATOR_SUFFIX);
		while(aJR.hasNext())
		{	Event event=aJR.next();
			if (event==Event.END_ARRAY)
			{	break;
			}
			else if (event==Event.START_ARRAY)
			{	fixNestedArrayNaming(aMutator,aName+ARRAY_INDICATOR_SUFFIX);
				parseHDSArray(aName,aJR,aMutator,aToString);
			}
			else if (event==Event.START_OBJECT)
			{	parseHDSObject(aName,aJR,aMutator,aToString);
			}
			else throw new IllegalArgumentException("unexpected event: "+event);
		}
		aMutator.popNode();
	}
	
	private static void fixNestedArrayNaming(IHDSMutator aMutator,String aName)
	{	//aMutator.g
		aMutator.rename(aName+ARRAY_INDICATOR_SUFFIX);
	}
	
	private static void parseHDSObject(String aName, JsonParser aJR, IHDSMutator aMutator, boolean aToString)
	{	if (aName!=null) aMutator.pushNode(aName);
		while(aJR.hasNext())
		{	Event event=aJR.next();
			if (event==Event.END_OBJECT)
			{	break;
			}
			if (event==Event.KEY_NAME)
			{	String name=aJR.getString();
				event=aJR.next();
				if (event==Event.START_OBJECT)
				{	parseHDSObject(name,aJR,aMutator,aToString);
				}
				else if (event==Event.START_ARRAY)
				{	parseHDSArray(name,aJR,aMutator,aToString);
				}
				else if (event==Event.VALUE_STRING)
				{	aMutator.addNode(name, aJR.getString());
				}
				else if (event==Event.VALUE_NULL)
				{	aMutator.addNode(name, null);
				}
				else if (event==Event.VALUE_NUMBER)
				{	if (aToString)
					{	aMutator.addNode(name, aJR.getString());
					}
					else
					{	aMutator.addNode(name, aJR.getBigDecimal());
					}
				}
				else if (event==Event.VALUE_FALSE)
				{	aMutator.addNode(name, false);
				}
				else if (event==Event.VALUE_TRUE)
				{	aMutator.addNode(name, true);
				}
				else throw new IllegalArgumentException("unexpected event: "+event);
			}
			else throw new IllegalArgumentException("unexpected event: "+event);
		}
		if (aName!=null) aMutator.popNode();
	}
	
	public static void main(String[] args)
	{
		try
		{	
			test("{}");
			test("[]");
			test("[[],[]]");
			test("[{},{}]");
			test("{\"a\":[[],[]]}");
			test("[{\"a\":2},{}]");
			test("{\"a\":\"abc\",\"b\":[{\"a\":2},{}]}");
			test("{\"a\":\"abc\",\"b\":1,\"c\":2.2,\"d\":null,\"e\":true}");
			test("{\"a\":\"abc\",\"a\":1}");
		}
		catch (Exception e)
		{	e.printStackTrace();
		}
	}
	
	private static void test(String aJSON)
	{
		//System.out.println(json);
		
		IHDSDocument hds=null;
		String json2=null;
		Exception ex=null;
		try
		{	hds=parseHDS(new ByteArrayInputStream(aJSON.getBytes()));
			//System.out.println(hds);
			ByteArrayOutputStream os=new ByteArrayOutputStream(1024);
			serializeHDS(os, hds,false);
			json2=new String(os.toByteArray());
		}
		catch (Exception e)
		{	ex=e;
		}
		
		if (json2!=null && json2.equals(aJSON))
		{	System.out.println("PASS: "+aJSON);
			System.out.println(hds);
		}
		else
		{	if (hds==null && ex!=null)
			{	System.out.println("FAIL(1): "+aJSON);
				ex.printStackTrace();
			}
			else if (json2==null && ex!=null)
			{	System.out.println("FAIL(2): "+aJSON);
				System.out.println(hds);
				ex.printStackTrace();
			}
			else 
			{	System.out.println("FAIL(3): "+aJSON+" -> "+json2);
				System.out.println(hds);
			}
		}
	}
	
}